/**************************************************************************
*  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  27.12.2017  mifi  First Version.
**************************************************************************/
#define __FPGA_C__

/*
 * The communication with the FPGA via SPI is not the optimal
 * implementation because these accesses are not atomic. Therefore,
 * access must take place under interrupt blocking. This is the only
 * way to ensure that an access in an interrupt works too.
 *
 * The following pins are used:
 *
 * SCK     P1.20
 * MISO    P1.23
 * MOSI    P1.24
 * CS      P1.31
 *
 * INTn    P0.21
 * RESET   P1.30   
 */  
  
/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/
#include <stdio.h>
#include <string.h>
#include "tal.h"
#include "fpga.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#define SPI_CS_LOW()    { LPC_GPIO1->CLR = (1u<<31); }
#define SPI_CS_HIGH()   { LPC_GPIO1->SET = (1u<<31); }

#define GPIO_INT_BIT    (1<<21)

#define FPGA_CMD_WRITE  0x20
#define FPGA_CMD_READ   0x40

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  IRQHandler                                                           */
/*                                                                       */
/*  This is our IRQ handler.                                             */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void IRQHandler (void)
{
   uint32_t dSource;

   /* Check if an interrupt is pending from PORT0 */
   if (Chip_GPIOINT_IsIntPending(LPC_GPIOINT, GPIOINT_PORT0))
   {
      /* Get source of falling edge interrupt from PORT0 */
      dSource = Chip_GPIOINT_GetStatusFalling(LPC_GPIOINT, GPIOINT_PORT0);
      
      /* If this our interrupt? */
      if (GPIO_INT_BIT == (dSource & GPIO_INT_BIT))
      {
         tal_LEDToggle(TAL_LED_CHANNEL_4);
      
         /* Clear TIMER interrupt */
         FPGAWrite(TIMER_STATUS, 1);
      }   
      
      /* Clear the interupt */
      Chip_GPIOINT_ClearIntStatus(LPC_GPIOINT, GPIOINT_PORT0, dSource);
   }
   
} /* IRQHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  GPIO_IRQHandler                                                      */
/*                                                                       */
/*  This is the Cortex GPIO IRQ handler.                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void GPIO_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler();
   TAL_CPU_IRQ_EXIT();
} /* GPIO_IRQHandler */

/*************************************************************************/
/*  FPGAInit                                                             */
/*                                                                       */
/*  Initialize the FPGA interface.                                       */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void FPGAInit (void)
{
   /* Turn on GPIO clock */
   Chip_Clock_EnablePeriphClock(SYSCTL_CLOCK_GPIO);

   /* 
    * Use SPI0, P1.20 (SCK), P1.23 (MISO), P1.24 (MOSI) and P1.31 (CS)  
    */

   /* Turn on SSP0 clock */
   Chip_Clock_EnablePeriphClock(SYSCTL_CLOCK_SSP0);

   /* Set GPIO P1.20, P1.23 and P1.24 to SSP0 mode */
   Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 20, (IOCON_FUNC5 | IOCON_MODE_INACT));  /*lint !e845*/
   Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 23, (IOCON_FUNC5 | IOCON_MODE_INACT));  /*lint !e845*/
   Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 24, (IOCON_FUNC5 | IOCON_MODE_INACT));  /*lint !e845*/


   /* Set GPIO 1.31 to GPIO mode */
   Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 31, (IOCON_FUNC0 | IOCON_DIGMODE_EN));
   
   /* Set GPIO to high */
   SPI_CS_HIGH();

   /* Set GPIO 1.31 to output */
   Chip_GPIO_SetPinDIROutput(LPC_GPIO, 1, 31);
   

   /*
    * Setup SSP0
    */
   
   /* Clock Prescale Register */
   LPC_SSP0->CPSR = 4;
   
   /* Control Register 0/1 */
   LPC_SSP0->CR0  = 7;  /* 8-bit transfer */
   LPC_SSP0->CR1  = SSP_CR1_SSP_EN;

   
   /*
    * Reset (P1.30)
    */
   Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 30, (IOCON_FUNC0 | IOCON_DIGMODE_EN));
   FPGAResetSet();
   Chip_GPIO_SetPinDIROutput(LPC_GPIO, 1, 30);
   FPGAResetClear();
   
   
   /*
    * Interrupt (P0.21)
    */
   Chip_IOCON_PinMuxSet(LPC_IOCON, 1, 21, (IOCON_FUNC0 | IOCON_MODE_PULLUP));
   Chip_GPIO_SetPinDIRInput(LPC_GPIO, 0, 21);
   Chip_GPIOINT_SetIntFalling(LPC_GPIOINT, GPIOINT_PORT0, GPIO_INT_BIT);
   
   /* Set priority and enable the IRQ */
   tal_CPUIrqSetPriority(GPIO_IRQn, 8);
   tal_CPUIrqEnable(GPIO_IRQn);

} /* FPGAInit */

/*************************************************************************/
/*  FPGAResetSet                                                         */
/*                                                                       */
/*  Assert reset signal which is low active.                             */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void FPGAResetSet (void)
{
   LPC_GPIO1->CLR = (1<<30);
} /* FPGAResetSet */

/*************************************************************************/
/*  FPGAResetClear                                                       */
/*                                                                       */
/*  Deassert reset signal which is high active.                          */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void FPGAResetClear (void)
{
   LPC_GPIO1->SET = (1<<30);
} /* FPGAResetClear */

/*************************************************************************/
/*  FPGATest                                                             */
/*                                                                       */
/*  Test the FPGA interface.                                             */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 0 = OK / -1 = ERROR                                          */
/*************************************************************************/
int FPGATest (void)
{
   int      nResult = -1;
   uint8_t  bIndex;
   uint8_t  bData;
   uint8_t   DataWrite[8]  = {1, 2, 3, 4, 5, 6, 7, 8};
   uint8_t   DataWrite2[8] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};
   uint8_t   DataWrite3[8] = {0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA};
   uint8_t   DataWrite4[8] = {0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55};
   uint8_t   Data[8];

   /*
    * A Test register at FPGA_TEST_BASE can be used for testing the interface.
    * This register contains 8 registers, which can be read and write.
    */
    
   /* 
    * Testing Write / Read 
    */
   
   /* Fill the register */   
   for (bIndex=0; bIndex<8; bIndex++)
   {
      FPGAWrite(FPGA_TEST_BASE + bIndex, bIndex+1);
   }   
   
   /* Check the content of the register */
   for (bIndex=0; bIndex<8; bIndex++)
   {
      bData = FPGARead(FPGA_TEST_BASE + bIndex);
      if (bData != (bIndex+1))
      {
         /* Error */
         goto FPGATestEnd; /*lint !e801*/
      }
   }
   
   
   /*
    * Write and Read ist working, testing WriteMulti
    */
    
   /* Fill the register first with 0 */   
   for (bIndex=0; bIndex<8; bIndex++)
   {
      FPGAWrite(FPGA_TEST_BASE+bIndex, 0);
   }   

   /* Write new data with WriteMuli */
   FPGAWriteMulti(TEST_REG_0, &DataWrite[0], 1);
   FPGAWriteMulti(TEST_REG_1, &DataWrite[1], 1);
   FPGAWriteMulti(TEST_REG_2, &DataWrite[2], 2);
   FPGAWriteMulti(TEST_REG_4, &DataWrite[4], 4);

   /* Check the content of the register */
   for (bIndex=0; bIndex<8; bIndex++)
   {
      bData = FPGARead(FPGA_TEST_BASE + bIndex);
      if (bData != (bIndex+1))
      {
         /* Error */
         goto FPGATestEnd; /*lint !e801*/
      }
   }
   
   
   /*
    * WriteMulti is working, testing ReadMulti
    */
    
   /* Fill the register first with 0 */   
   for (bIndex=0; bIndex<8; bIndex++)
   {
      FPGAWrite(FPGA_TEST_BASE + bIndex, 0);
   }   
   
   /* Write new data with WriteMuli */
   FPGAWriteMulti(TEST_REG_0, DataWrite, 8);
   
   /* Read data with ReadMulti */
   memset(Data, 0x00, sizeof(Data));
   
   FPGAReadMulti(TEST_REG_0, &Data[0], 1);
   FPGAReadMulti(TEST_REG_1, &Data[1], 1);
   FPGAReadMulti(TEST_REG_2, &Data[2], 2);
   FPGAReadMulti(TEST_REG_4, &Data[4], 4);
   
   /* Check the values */
   if (memcmp(DataWrite, Data, 8) != 0)
   {
      /* Error */
      goto FPGATestEnd; /*lint !e801*/
   }
   
   
   /*
    *  WriteMulti test #1
    */

   /* Write new data with WriteMuli */
   FPGAWriteMulti(TEST_REG_0, DataWrite, 8);
   
   /* Read data with ReadMulti */
   memset(Data, 0x00, sizeof(Data));

   FPGAReadMulti(TEST_REG_4, &Data[4], 4);
   FPGAReadMulti(TEST_REG_0, &Data[0], 4);
   
   /* Check the values */
   if (memcmp(DataWrite, Data, 8) != 0)
   {
      /* Error */
      goto FPGATestEnd; /*lint !e801*/
   }


   /*
    *  WriteMulti test #2
    */
   
   /* Write new data with WriteMuli */
   FPGAWriteMulti(TEST_REG_0, DataWrite2, 8);
   
   /* Read data with ReadMulti */
   memset(Data, 0x00, sizeof(Data));

   FPGAReadMulti(TEST_REG_0, &Data[0], 4);
   FPGAReadMulti(TEST_REG_4, &Data[4], 4);
   
   /* Check the values */
   if (memcmp(DataWrite2, Data, 8) != 0)
   {
      /* Error */
      goto FPGATestEnd; /*lint !e801*/
   }


   /*
    *  WriteMulti test #3
    */

   /* Write new data with WriteMuli */
   FPGAWriteMulti(TEST_REG_0, DataWrite3, 8);
   
   /* Read data with ReadMulti */
   memset(Data, 0x00, sizeof(Data));

   FPGAReadMulti(TEST_REG_0, &Data[0], 4);
   FPGAReadMulti(TEST_REG_4, &Data[4], 4);
   
   /* Check the values */
   if (memcmp(DataWrite3, Data, 8) != 0)
   {
      /* Error */
      goto FPGATestEnd; /*lint !e801*/
   }
   
   
   /*
    *  WriteMulti test #4
    */
      
   /* Write new data with WriteMuli */
   FPGAWriteMulti(TEST_REG_0, DataWrite4, 8);
   
   /* Read data with ReadMulti */
   memset(Data, 0x00, sizeof(Data));

   FPGAReadMulti(TEST_REG_0, &Data[0], 4);
   FPGAReadMulti(TEST_REG_4, &Data[4], 4);
   
   /* Check the values */
   if (memcmp(DataWrite4, Data, 8) != 0)
   {
      /* Error */
      goto FPGATestEnd; /*lint !e801*/
   }

   
   /*
    * Write, Read, WriteMulti and ReadMulti is working now.
    * But the WriteMulti can write up to 255 bytes, and
    * ReadMulti can read up to 4 bytes.
    */
    
    
   /* 
    * Enable timer for interrupt test
    */ 
   FPGATimerInit(1000);
   FPGATimerStart();
   
   nResult = 0;
   
FPGATestEnd:   
   return(nResult);
} /* FPGATest */

/*************************************************************************/
/*  FPGARead                                                             */
/*                                                                       */
/*  Read one byte from the FPGA address.                                 */
/*                                                                       */
/*  In    : bAddress                                                     */
/*  Out   : none                                                         */
/*  Return: bData                                                        */
/*************************************************************************/
uint8_t FPGARead (uint16_t wAddress)
{
   uint32_t dDummy;
   uint8_t  bData;
   uint8_t  bAdrHi;
   uint8_t  bAdrLo;
   
   TAL_CPU_DISABLE_ALL_INTS();

   /* Setup command and address */   
   bAdrHi = FPGA_CMD_READ | (wAddress >> 8);
   bAdrLo = (wAddress & 0x00FF);

   /* Write command, address and data */
   SPI_CS_LOW();
   LPC_SSP0->DR = bAdrHi;
   LPC_SSP0->DR = bAdrLo;
   LPC_SSP0->DR = 0;       /* Count - 1 */
   LPC_SSP0->DR = 0xFF;    /* Dummy */

   /* Wait while SSP is busy */
   while (LPC_SSP0->SR & SSP_STAT_BSY)
   {
      __asm__ volatile ("nop");      
   }

   SPI_CS_HIGH();

   /* Get data */   
   dDummy = LPC_SSP0->DR;   
   dDummy = LPC_SSP0->DR;  /*lint !e838*/   
   dDummy = LPC_SSP0->DR;  /*lint !e838*/ 
   (void)dDummy;
   
   bData = (uint8_t)LPC_SSP0->DR;
   
   TAL_CPU_ENABLE_ALL_INTS();   

   return(bData);
} /* FPGARead */

/*************************************************************************/
/*  FPGAWrite                                                            */
/*                                                                       */
/*  Write one byte to the FPGA address.                                  */
/*                                                                       */
/*  In    : wAddress, bData                                              */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void FPGAWrite (uint16_t wAddress, uint8_t bData)
{
   uint32_t dDummy;
   uint8_t  bAdrHi;
   uint8_t  bAdrLo;

   TAL_CPU_DISABLE_ALL_INTS();

   /* Setup command and address */   
   bAdrHi = FPGA_CMD_WRITE | (wAddress >> 8);
   bAdrLo = (wAddress & 0x00FF);

   /* Write command, address and data */
   SPI_CS_LOW();
   LPC_SSP0->DR = bAdrHi;
   LPC_SSP0->DR = bAdrLo;
   LPC_SSP0->DR = bData;

   /* Wait while SSP is busy */
   while (LPC_SSP0->SR & SSP_STAT_BSY)
   {
      __asm__ volatile ("nop");      
   }

   SPI_CS_HIGH();
   
   /* Empty FIFO */
   while (LPC_SSP0->SR & SSP_STAT_RNE)
   {
      dDummy = LPC_SSP0->DR;
      (void)dDummy;
   }

   TAL_CPU_ENABLE_ALL_INTS();   

} /* FPGAWrite */

/*************************************************************************/
/*  FPGAReadMulti                                                        */
/*                                                                       */
/*  Read multiple bytes from the FPGA.                                   */
/*  The address will be auto incremented by the FPGA.                    */
/*                                                                       */
/*  In    : bAddress, pBuffer, bCount                                    */
/*  Out   : none                                                         */
/*  Return: bData                                                        */
/*************************************************************************/
void FPGAReadMulti (uint16_t wAddress, void *pBuffer, uint8_t bCount)
{
   uint8_t *pData = (uint8_t*)pBuffer;
   uint32_t dDummy;
   uint8_t  bAdrHi;
   uint8_t  bAdrLo;
   uint8_t  bIndex;
   
   if (bCount <= 4)
   {
      TAL_CPU_DISABLE_ALL_INTS();
   
      /* Setup command and address */   
      bAdrHi = FPGA_CMD_READ | (wAddress >> 8);
      bAdrLo = (wAddress & 0x00FF);

      /* Write command, address and data */
      SPI_CS_LOW();
      LPC_SSP0->DR = bAdrHi;
      LPC_SSP0->DR = bAdrLo;
      LPC_SSP0->DR = bCount-1;   /* Count - 1 */
   
      bIndex = bCount;
      while(bIndex--)
      {
         LPC_SSP0->DR = 0xFF;    /* Dummy */
      }   

      /* Wait while SSP is busy */
      while (LPC_SSP0->SR & SSP_STAT_BSY)
      {
         __asm__ volatile ("nop");      
      }

      SPI_CS_HIGH();

      /* Get data */   
      dDummy = LPC_SSP0->DR;   
      dDummy = LPC_SSP0->DR;  /*lint !e838*/   
      dDummy = LPC_SSP0->DR;  /*lint !e838*/ 
      (void)dDummy;
   
      while(bCount--)
      {
         *pData++ = (uint8_t)LPC_SSP0->DR;
      }
      
      TAL_CPU_ENABLE_ALL_INTS();
   }      
   
} /* FPGAReadMulti */

/*************************************************************************/
/*  FPGAWriteMulti                                                       */
/*                                                                       */
/*  Write multiple bytes to the FPGA.                                    */
/*  The address will be auto incremented by the FPGA.                    */
/*                                                                       */
/*  In    : wAddress, *pBuffer, bCount                                   */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void FPGAWriteMulti (uint16_t wAddress, void *pBuffer, uint8_t bCount)
{
   uint8_t *pData = (uint8_t*)pBuffer;
   uint32_t dDummy;
   uint8_t  bAdrHi;
   uint8_t  bAdrLo;

   TAL_CPU_DISABLE_ALL_INTS();

   /* Setup command and address */   
   bAdrHi = FPGA_CMD_WRITE | (wAddress >> 8);
   bAdrLo = (wAddress & 0x00FF);

   /* Write command and address */
   SPI_CS_LOW();
   LPC_SSP0->DR = bAdrHi;
   LPC_SSP0->DR = bAdrLo;

   /* Write data */
   switch (bCount)
   {
      case 1:
      {
         LPC_SSP0->DR = *pData;
         break;
      }
   
      case 2:
      {
         LPC_SSP0->DR = *pData++;
         LPC_SSP0->DR = *pData;
         break;
      }
   
      case 4:
      {
         LPC_SSP0->DR = *pData++;
         LPC_SSP0->DR = *pData++;
         LPC_SSP0->DR = *pData++;
         LPC_SSP0->DR = *pData;
         break;
      }
      
      default:
      {
         while (bCount--)
         {
            LPC_SSP0->DR = *pData++;
         }
         break;
      }
   }
   
   /* Wait while SSP is busy */
   while (LPC_SSP0->SR & SSP_STAT_BSY)
   {
      __asm__ volatile ("nop");      
   }

   SPI_CS_HIGH();
   
   /* Empty FIFO */
   while (LPC_SSP0->SR & SSP_STAT_RNE)
   {
      dDummy = LPC_SSP0->DR;
      (void)dDummy;
   }
   
   TAL_CPU_ENABLE_ALL_INTS();

} /* FPGAWriteMulti */

/*************************************************************************/
/*  FPGATimerInit                                                        */
/*                                                                       */
/*  Configure the timer.                                                 */
/*                                                                       */
/*  In    : wMilliSec                                                    */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void FPGATimerInit (uint16_t wMilliSec)
{
   uint32_t Value;

   Value = ((FPGA_FRQ_HZ / 1000) * wMilliSec) - 1;
   
   FPGAWrite(TIMER_CONTROL, 0);  /* Disable timer */
   FPGAWrite4(TIMER_PERIOD_0, &Value);
   
} /* FPGATimerInit */

/*************************************************************************/
/*  FPGATimerStart                                                       */
/*                                                                       */
/*  Start the timer.                                                     */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void FPGATimerStart (void)
{
   FPGAWrite(TIMER_CONTROL, 1);
} /* FPGATimerStart */

/*************************************************************************/
/*  FPGATimerStop                                                        */
/*                                                                       */
/*  Stop the timer.                                                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void FPGATimerStop (void)
{
   FPGAWrite(TIMER_CONTROL, 0);
} /* FPGATimerStop */

/*** EOF ***/
